# http binding interface
# $Id: rest.py,v 1.2 2005/03/10 00:39:10 tofu Exp $
from twisted.python import components
from twisted.web import server, resource
from twisted.internet import reactor, protocol, defer
from twisted.xish import domish, utility

import sha, base64
import session
import types

# make REST body class, similar to xmlrpclib
# 
class RESTParse:
    def __init__(self):
        # Setup the parser
        self.stream = domish.elementStream()
        self.stream.DocumentStartEvent = self.onDocumentStart
        self.stream.ElementEvent = self.onElement
        self.stream.DocumentEndEvent = self.onDocumentEnd
        self.body = ""
        self.elements = []
        

    def parse(self,buf):
        self.stream.parse(buf)
        return self.body,self.elements
    
    def serialize(self, obj):
        if isinstance(obj, domish.Element):
            obj = obj.toXml()
        return obj

    def onDocumentStart(self, rootelem):
        if rootelem.name == 'body':
            self.body = rootelem

    def onElement(self, element):
        if isinstance(element, domish.Element):
            self.elements.append(element)
        else:
            pass


    def onDocumentEnd(self):
        pass

    
class REST(resource.Resource):
    isLeaf = True
    def __init__(self,v=0):
        """Initialize.
        """
        self.children = {}
        self.sessions = {}
        self.counter  = 0
        self.client   = 0
        self.v        = v
        
    def render_GET(self, request):
        return "<html><body><a href='http://www.jabber.org/jeps/jep-0124.html'>JEP-0124</a> - HTTP Binding</body></html>"
    def render_POST(self, request):
        new_session = 0       
        request.content.seek(0, 0)       
        hp = RESTParse()
        try:
            body_tag,elements = hp.parse(request.content.read())            
            if body_tag.name != "body":
                self.send_http_error(400,request)
                return server.NOT_DONE_YET            
        except:
            
            self.send_http_error(400,request) 
            return server.NOT_DONE_YET
        else:             
            # check for existing session
            if body_tag.hasAttribute('sid'):
                if self.sessions.has_key(body_tag['sid']):
                    s = self.sessions[body_tag['sid']]
                    # check for keys
                    # TODO - clean this up
                    
                    if body_tag.hasAttribute('newkey'):
                        newkey = body_tag['newkey']
                        s.key = newkey
                    try:
                        if body_tag.hasAttribute('key'):
                            if s.key is not None:
                                nk = sha.new(body_tag['key'])
                                key = nk.hexdigest()
                                next_key = body_tag['key']
                                if key == s.key:
                                    s.key = next_key
                                else:
                                    self.send_http_error(401,request)
                                    return server.NOT_DONE_YET
                                
                            else:
                                self.send_http_error(401,request)
                                return server.NOT_DONE_YET
                        
                    except:
                        self.send_http_error(401,request)
                        return server.NOT_DONE_YET
                    
                    # need to check if this is a valid rid
                    if body_tag.hasAttribute('rid') and body_tag['rid']!='': 
                        if int(body_tag['rid']) != int(s.rid):
                            self.send_http_error(401,request,'internal-server-error')
                            return server.NOT_DONE_YET
                    else:
                        self.send_http_error(401,request,'internal-server-error')
                        return server.NOT_DONE_YET
                else:
                    self.send_http_error(403,request,'internal-server-error')
                    return server.NOT_DONE_YET
            else:
                new_session = 1                                
                # look for rid
                if body_tag.hasAttribute('rid') and body_tag['rid']!='':
                    rid = body_tag['rid']
                else:
                    self.send_http_error(401,request,'internal-server-error')
                    return server.NOT_DONE_YET

                # look for to
                if body_tag.hasAttribute('to') and body_tag['to']!='':
                    to = body_tag['to']
                    
                else:
                    self.send_http_error(400,request,'internal-server-error')
                    return server.NOT_DONE_YET
                
                # look for wait
                if body_tag.hasAttribute('wait') and body_tag['wait']!='':
                    wait = body_tag['wait']
                else:
                    self.send_http_error(400,request,'internal-server-error')
                    return server.NOT_DONE_YET


                # look for lang
                
                if body_tag.hasAttribute("xml:lang") and body_tag['xml:lang']!='':
                    lang = body_tag.getAttribute("xml:lang")
                else:
                    for k in body_tag.attributes:
                        if isinstance(k, tuple):
                            if str(k[1])=='lang' and body_tag.getAttribute(k) !='':
                                lang = body_tag.getAttribute(k)
                            else:
                                self.send_http_error(400,request,'internal-server-error')
                                return server.NOT_DONE_YET

                d = session.make_session(self,body_tag.attributes)

            # process session events and body tag
            if new_session == 0:
                try:
                    # grab session                    
                    if body_tag.hasAttribute('sid'):
                        sid = str(body_tag['sid'])
                    if body_tag.hasAttribute('to') and body_tag['to']!='':
                        self.send_http_error(400,request,'internal-server-error')
                        return server.NOT_DONE_YET
                    
                    if self.sessions.has_key(sid):
                        s = self.sessions[sid]
                        
                        d = self.parseBody(s, body_tag, elements,request)
                        if d == 0:
                            self.return_REST(s,request)
                        else:
                            d.addCallback(self.return_REST, request)
                            d.addErrback(self.return_error, request) 
                        
                    else:
                        self.send_http_error(401,request,'internal-server-error')
                        return server.NOT_DONE_YET
                except:
                    self.send_http_error(401,request,'internal-server-error')
                    return server.NOT_DONE_YET
            else:
                d.addCallback(self.return_session, request)
                d.addErrback(self.return_error, request)
                
        return server.NOT_DONE_YET
        

    def parseBody(self, session, body_tag, elements, request):        
        self.waiting = 0
        session.rid  = session.rid + 1
        d = 0
        for i,v in enumerate(elements):
            if v.name == 'iq':
                self.waiting = self.waiting + 1
                            
            d = session.sendXml(v)
        if body_tag.hasAttribute('type'):
            if body_tag['type'] == 'terminate':
                session.wait = 0
                session.expire()
                return 0      
        
        return d

    def return_session(self,session, request):
        # create body
        b = domish.Element(("http://jabber.org/protocol/httpbind","body"))       
        if session.authid != 0:
            b['authid'] = session.authid
        b['sid']    = session.sid
        b['xmlns']  = "http://jabber.org/protocol/httpbind"
        if session.attrs.has_key('content'):
            b['content'] = session.attrs['content']

        self.return_body(request,b)

    def return_REST(self, obj, request):
        xml = None
        if isinstance(obj, types.TupleType):
            session,xml = obj
        else:
            session = obj
        # create body
        b = domish.Element(("http://jabber.org/protocol/httpbind","body"))
        
        b['xmlns']  = "http://jabber.org/protocol/httpbind"
        if session.attrs.has_key('content'):
            b['content'] = session.attrs['content']
        if xml is not None:
            b.addChild(xml)
        # check for messages, presence, and iq
        if xml is None or not xml.hasAttribute('id'):
            #reactor.callLater(session.wait,self.wait_xml,b,session,request)
            reactor.callLater(1,self.wait_xml,b,session,request)
            return server.NOT_DONE_YET
                
        self.return_body(request,b)        

    def wait_xml(self,b,session,request):
        for i,obj in enumerate(session.msgs):
            m = session.msgs.pop(0)
            b.addChild(m)            
        for i,obj in enumerate(session.prs):
            p = session.prs.pop(0)
            b.addChild(p)            
        for i,obj in enumerate(session.iqs):
            iq = session.iqs.pop(0)
            b.addChild(iq)
        self.return_body(request,b)        
        

    def return_error(self,error, request):
        self.send_http_error(500,request,'internal-server-error','error',error)
        

    def return_body(self,request,b):        
        request.setResponseCode(200)
        request.setHeader('content-type','text/xml')
        request.setHeader('content-length', str(len(b.toXml())))
        request.write(b.toXml())
        request.finish()
            
    def send_http_error(self,code,request, condition = 'undefined-condition', type = 'terminate', data = ''):
        request.setResponseCode(code)
        
        b = domish.Element(("http://jabber.org/protocol/httpbind","body"))
        b['condition'] = condition
        b['type']   = type
        if code != 400:
            if data != '':
                if condition == 'see-other-uri':
                    b.addElement('uri','',content = data)
                else:
                    b.addElement('text','urn:ietf:params:xml:ns:xmpp-streams',content = data)
            request.setHeader("content-length", len(b.toXml()))    
            request.write(b.toXml())
        else:
            request.setHeader("content-length", "0")
        request.finish()
