from twisted.python import components, log
from twisted.web import server, resource, xmlrpc
from twisted.internet import reactor, protocol, defer

try:
    from twisted.words.xish import domish, utility
except:
    from twisted.xish import domish, utility


import xmlrpclib
import random 
import sha
import session

class XMLRPCPJab(xmlrpc.XMLRPC):
    """ An xmlrpc interface to punjab jabber connections
    """

    
    def __init__(self,v=0):
        xmlrpc.XMLRPC.__init__(self)
        self.counter  = 0
        self.sessions = {}
        self.v        = v
    
    def xmlrpc_connect(self, jid, passwd, port=5222, rid = random.randint(10000,1000000)):
        """ Connect to a jabber server """
        attrs = {}
        try:
            attrs['user'],attrs['to'] = jid.split('@')
        except:
            attrs['user'] = jid
            attrs['to']   = jid
            
        if attrs['to'].find('/') != -1:
            attrs['to'], attrs['resource'] = attrs['to'].split('/')
        else:
            attrs['resource'] = "PunJab"
            
        attrs['rid'] = rid
        attrs['wait'] = 0
        attrs['hold'] = 0
        attrs['xml:lang'] = 'en' # this is english for now
        
        s, d = session.make_session(self,attrs)              
        d.addCallback(self.login, s, jid, passwd)
        d.addErrback(self.return_error)
        return d # this needs to be cleaned up, I need to understand this better

    xmlrpc_connect.signature = [
        ['string', 'string'],
        ['string', 'string', 'string','int'],
        # TODO - add keys like httpb
        ]
    xmlrpc_connect.help = "Given a jabber id and password you can connect to a jabber server."


    def xmlrpc_is_connected(self, sid):
        """ Check if a punjab session is connected to a jabber server. This may be taken away. Insecure?"""
        if self.sessions.has_key(sid):
            return sid
        else:
            return 0
    xmlrpc_is_connected.signature = [
        ['string'],
        # TODO - add keys like httpb
        ]


    def login(self, data, session, jid, passwd):
        # login to the jabber server        
        d = session.startLogin(jid,passwd)
        d.addCallback(self.return_login)
        d.addErrback(self.return_error)
        return d

    def return_login(self, session):
        # maybe do some other stuff?
        return session.sid

    def xmlrpc_disconnect(self,sid):
        """ disconnect from the jabber server and close your session. """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]
                s.terminate()
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in disconnect.")
        
        return 1
    xmlrpc_disconnect.signature = [
        ['string'],
        # TODO - add keys like httpb
        ]
    xmlrpc_disconnect.help = "Given a punjab session id you will disconnect from a jabber server."


    def xmlrpc_start_contact(self, sid, service, to, nick = '', group = '', request = 0):
        """ Subscribe to a jabber entities presence.
        """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]        
                if request == 1:
                    s.sendPresence(to,'subscribe','Online','Online')

                d = s.setRoster(to,nick,group)
                d.addCallback(self.return_send)
                d.addErrback(self.return_error)
                return d
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in start contact")
    xmlrpc_start_contact.signature = [
            ['string', 'string', 'string'],
            ['string', 'string', 'string','string','string','int'],
            # TODO - add keys like httpb
                                ]
    xmlrpc_start_contact.help = "Given a punjab session id, a jabber id, nickname, and grouping you  initiate a subscription request to the jabber id."

        

    def xmlrpc_start_roster(self, sid):
        """ Request your jabber roster.
        """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]        
            
                d = s.startRoster()
                d.addCallback(self.return_obj)
                d.addErrback(self.return_error)
                return d
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            raise # if debug then raise
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in roster. ")

    xmlrpc_get_roster = xmlrpc_start_roster

    xmlrpc_get_roster.signature = [
            ['string'],
            # TODO - add keys like httpb
            ]
    xmlrpc_get_roster.help = "Grab your current roster that is located on the jabber server your punjab client is connected to."

    xmlrpc_start_roster.signature = [
        ['string'],
        # TODO - add keys like httpb
        ]
    xmlrpc_start_roster.help = "Grab your current roster that is located on the jabber server your punjab client is connected to."

    

    def xmlrpc_send_presence(self, sid, to = '', type = '', show = '', status = '', priority = 0):
        """ Send a presence packet to a jabber entity.
        """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]        
                s.sendPresence(to,type,show,status,priority)
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in send presence. ")
        
        return 1
    xmlrpc_send_presence.signature = [
        ['string'],
        ['string', 'string', 'string','string','string','int'],
        # TODO - add keys like httpb
        ]
    xmlrpc_send_presence.help = "Given a punjab session id, an option to jabber id, and option presence type, and optional show, an option status, and an optional priority level you can send presence to a jabber server or the given to jabber id."


    def return_send(self, obj):
        return 1
        
    def return_obj(self, obj):
        return obj

    def return_error(self, error):
        return xmlrpc.Fault(xmlrpc.FAILURE, "error : " + str(error))
        
    error = return_error
    
    def xmlrpc_get_presence(self, sid, to = '', f = ''):
        """ Get a presence packet that has been sent to you.
        """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]        
                
                return s.getPresence(to,f)
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            raise # if debug then raise
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in get presence.")
        
        return 0

    xmlrpc_get_presence.signature = [
        ['string', 'string', 'string'],
        # TODO - add keys like httpb
        ]
    xmlrpc_get_presence.help = "Given a punjab session id, an optional to jabber id, and an optional from jabber id you can query presence elements that have been sent to the punjab client connection. "


    def xmlrpc_send_message(self, sid, to, subject = '', body = '', type = 'chat', id = 0):
        """ Send a message packet to a jabber entity.
        """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]        
                s.sendMessage(to,subject,body,type,id)
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in send message. ")
        
        return 1
    xmlrpc_send_message.signature = [
        ['string', 'string', 'string','string','string','string','string'],
        ['string', 'string', 'string','string','string','string','int'],
        # TODO - add keys like httpb
        ]
    xmlrpc_send_message.help = "Given a punjab session id, a to jabber id, an optional subject, a body, an optional message type, and an option thread id you can send messages to a jabber id."


    def xmlrpc_get_message(self, sid, to = '', type = '', id = 0, f = ''):
        """ get a message that has been sent to you and deletes it from punjab.
        """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]        
            
                return s.getMessage(to,f, type, id)
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            raise # if debug then raise
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in get_message.")
        
        return 0
    xmlrpc_get_message.signature = [
        ['string', 'string', 'string','string','string'],
        ['string', 'string', 'string','int','string'],
        # TODO - add keys like httpb
        ]
    xmlrpc_get_message.help = "Given a session id you can check if messages exist in the punjab message queue. Add to, type, thread, from to query specific messages that have been sent to the client connection. Deletes message from queue."



    def xmlrpc_check_message(self, sid, to = '', type = '', id = 0, f = ''):
        """ get a message that has been sent to you and DO NOT delete it from punjab.
        """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]        
                
                return s.checkMessage(to,f, type, id)
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            raise # if debug then raise
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in check message. ")
        
        return 0

    xmlrpc_check_message.signature = [
        ['string', 'string', 'string','string','string'],
        ['string', 'string', 'string','int','string'],
        # TODO - add keys like httpb
        ]
    xmlrpc_check_message.help = "Given a session id you can check if messages exist in the punjab message queue. Add to, type, thread, from to query specific messages that have been sent to the client connection. Does not delete message from queue."


    def xmlrpc_get_iq(self, sid, id):
        """ get iq packets that have been sent to you.
        """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]        
                
                return s.getIq(id)

            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            raise # if debug then raise
            return xmlrpc.Fault(xmlrpc.FAILURE, " Error in get_iq. ")
        
        return 0

    xmlrpc_get_iq.signature = [
        ['string', 'string'],
        # TODO - add keys like httpb
        ]
    xmlrpc_get_iq.help = "Given a punjab session id and a iq element id you can grab iq elements that have been send to your client connection."


    def xmlrpc_browse(self, sid, jid = '', auto = 0, items = 0):
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]                                
                                
                d = s.browse(jid,auto,items)                
                d.addCallback(self.return_obj)
                d.addErrback(self.return_error)
                return d 
                
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            raise
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in browse. ")
        
        return 0
    xmlrpc_browse.signature = [
        ['string', 'string', 'int','int'],
        # TODO - add keys like httpb
        ]
    xmlrpc_browse.help = "Given a punjab session id, a jabber id you can use the legacy jabber browse mechanism and/or disco to query features or items the given jabber id contains. If you leave the jabber id blank it will query the sever the session id is connected to."



    def xmlrpc_vcard(self, sid, jid):
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]        
                
                d = s.vcard(jid)
                d.addCallback(self.return_obj)
                d.addErrback(self.return_error)
                return d 
            
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            raise # if debug then raise
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in vcard ")
        
        return 0
    xmlrpc_vcard.signature = [
        ['string', 'string'],
        # TODO - add keys like httpb
        ]
    xmlrpc_vcard.help = "Given a punjab session id and a jabber id you can get a vcard of the given jabber id."


        

    def xmlrpc_send_xml(self, sid, xml):
        """ send any xml to the jabber server
        """
        try:
            if self.sessions.has_key(sid):
                s = self.sessions[sid]                        
                d = s.sendXml(xml)
                d.addCallback(self.return_obj)
                d.addErrback(self.return_error)
                return d 
            
            else:
                return xmlrpc.Fault(xmlrpc.FAILURE, "Session Does not exists. Perhaps you are not logged in? ")
        except:
            return xmlrpc.Fault(xmlrpc.FAILURE, "Error in sending xml. ")
    
    xmlrpc_send_xml.signature = [
        ['string', 'string'],
        # TODO - add keys like httpb
        ]
    xmlrpc_send_xml.help = "Given an xml string and session id you can send that xml to the jabber server."
