# twonds component

from zope.interface import implements
try:
    from twisted.words.xish import domish, xpath, utility
    from twisted.words.protocols.jabber import xmlstream
except:
    from twisted.xish import domish, xpath, utility, xmlstream
    
from twisted.words.protocols.jabber import jstrports, jid

import client


def componentFactory(componentid, password, id = None):
    a = ConnectComponentAuthenticator(componentid, password, id)
    return xmlstream.XmlStreamFactory(a)


COMP_BIND_XMLNS = 'http://jabberd.jabberstudio.org/ns/component/1.0'

class ConnectComponentAuthenticator(client.XMPPAuthenticator):
    """ Authenticator to permit an XmlStream to authenticate against a Jabber
    Server as a Component (where the Authenticator is initiating the stream).

    This implements the basic component authentication. Unfortunately this
    protocol is not formally described anywhere. Fortunately, all the Jabber
    servers I know of use this mechanism in exactly the same way.

    """
    namespace = 'jabber:component:accept'
    version = '1.0'
    
    def __init__(self, componentjid, password, id = None):
        """
        @type componentjid: L{str}
        @param componentjid: Jabber ID that this component wishes to bind to.

        @type password: L{str}
        @param password: Password/secret this component uses to authenticate.
        """

        if isinstance(componentjid,str):
            
            componentjid = jid.JID(componentjid)
            if componentjid.user is None:
                 componentjid.user = componentjid.host
                 componentjid.host = ''
                 
            
        client.XMPPAuthenticator.__init__(self, componentjid, password)
        self.password = password
        self.id       = id

    def connectionMade(self):
        # Generate stream header
        if self.version != 0.0:
            sh = "<stream:stream  xmlns:stream='http://etherx.jabber.org/streams' version='%s' >" % \
                 (self.version)
        else:
            sh = "<stream:stream xmlns='%s' xmlns:stream='http://etherx.jabber.org/streams' >" % \
                 (self.namespace)
        self.xmlstream.send(sh)

    def _proceed(self, p):
        from twisted.internet import ssl
        # Reconnect using SSL
        ctx = ssl.ClientContextFactory()
        self.xmlstream.transport.startTLS(ctx)
        self.xmlstream._reset()
        # Generate stream header
        if self.version != 0.0:
            sh = "<stream:stream  xmlns:stream='http://etherx.jabber.org/streams' version='%s' >" % \
                 (self.version)
        else:
            sh = "<stream:stream xmlns='%s' xmlns:stream='http://etherx.jabber.org/streams' >" % \
                 (self.namespace)
        self.xmlstream.send(sh)
        self.tls = 1

    def streamStarted(self, rootelem):
        if rootelem.hasAttribute('version'):
            self.version = rootelem['version']

        if self.version == 0.0:
            # Create handshake
            hs = domish.Element(("jabber:component:accept", "handshake"))
            hs.addContent(xmlstream.hashPassword(self.xmlstream.sid, self.password))
            
            # Setup observer to watch for handshake result
            self.xmlstream.addOnetimeObserver("/handshake", self._handshakeEvent)
            self.xmlstream.send(hs)
        else:
            self.xmlstream.addOnetimeObserver("/features",self._featureParse)


    def _bindResult(self, b):
        # TODO - check for error
        self.bind = 0
        self.success = 0
        self.xmlstream.dispatch(self.xmlstream, xmlstream.STREAM_AUTHD_EVENT)
                
            
    def _featureParse(self, f):
        self.bind    = 0
        self.session = 0
        # TODO - check for tls
        if self.success == 1:
            for f in f.elements():
                if f.name == "bind":
                    self.bind = 1
                if f.name == "session":
                    self.session = 1
                    
            #if self.bind == 1:
                
            bind = domish.Element((client.BIND_XMLNS, "bind"))
            bind['xmlns'] = COMP_BIND_XMLNS
            bind['name']  = self.id
            
            self.xmlstream.addOnetimeObserver("/bind",self._bindResult)
            self.xmlstream.send(bind)
                
        else:
            if f.starttls:
                if SSL:
                    # look for required
                    starttls = domish.Element((client.TLS_XMLNS,"starttls"),TLS_XMLNS)
                    # why?
                    starttls['xmlns'] = TLS_XMLNS
                    self.xmlstream.addOnetimeObserver("/proceed",self._proceed)
                    self.xmlstream.addOnetimeObserver("/failue",self._tlsError)
                    self.xmlstream.send(starttls)
                else:
                    self.xmlstream.dispatch(f, self.AUTH_FAILED_EVENT)
            else:
                # Look for SASL
                m = f.mechanisms
                
                if m.uri == client.SASL_XMLNS:
                    ms = 'DIGEST-MD5'
                    for mech in m.elements():
                        ms = str(mech)
                        if ms == 'DIGEST-MD5':
                            break
                        if ms == 'PLAIN':
                            break
                    auth = domish.Element((client.SASL_XMLNS,"auth"),client.SASL_XMLNS,{'mechanism' : ms})
                    # why?
                    auth['xmlns'] = client.SASL_XMLNS
                    # auth['mechanism'] = ms
                    if ms == 'DIGEST-MD5':
                        self.xmlstream.addOnetimeObserver("/challenge",self._saslStep1)
                    if ms == 'PLAIN':
                        # TODO add authzid
                        auth_str = ""
                        auth_str = auth_str + "\000"
                        auth_str = auth_str + self.jid.user.encode('utf-8')
                        auth_str = auth_str + "\000"
                        auth_str = auth_str + self.password.encode('utf-8')
                        auth.addContent(binascii.b2a_base64(auth_str))
                        self.xmlstream.addOnetimeObserver("/success",self._saslSuccess)
                        
                    self.xmlstream.addObserver("/failure",self._saslError)
                    self.xmlstream.send(auth)
                else:
                    self.xmlstream.dispatch(f, self.AUTH_FAILED_EVENT)

    def _handshakeEvent(self, elem):
        self.xmlstream.dispatch(self.xmlstream, xmlstream.STREAM_AUTHD_EVENT)

class ListenComponentAuthenticator(xmlstream.Authenticator):
    """ Placeholder for listening components """
    pass


from twisted.application import service
from zope.interface import Interface, implements

class IService(Interface):
    
    def componentConnected(self, xmlstream):
        """ Parent component has established a connection
        """

    def componentDisconnected(self):
        """ Parent component has lost a connection to the Jabber system
        """

    def transportConnected(self, xmlstream):
        """ Parent component has established a connection over the underlying transport
        """

class Service(service.Service):
    implements(IService)

    def componentConnected(self, xmlstream):
        pass

    def componentDisconnected(self):
        pass

    def transportConnected(self, xmlstream):
        pass

    def send(self, obj):
        self.parent.send(obj)


class ServiceManager(service.MultiService):
    """ Business logic representing a managed component connection to a Jabber router

    This Service maintains a single connection to a Jabber router and
    provides facilities for packet routing and transmission. Business
    logic modules can 
    subclasses, and added as sub-service.
    """
    def __init__(self, jid, password, id = None):
        service.MultiService.__init__(self)

        # Setup defaults
        self.jabberId = jid
        self.xmlstream = None

        # Internal buffer of packets
        self._packetQueue = []

        # Setup the xmlstream factory
        self._xsFactory = componentFactory(self.jabberId, password, id)

        # Register some lambda functions to keep the self.xmlstream var up to date
        self._xsFactory.addBootstrap(xmlstream.STREAM_CONNECTED_EVENT, self._connected)
        self._xsFactory.addBootstrap(xmlstream.STREAM_AUTHD_EVENT, self._authd)
        self._xsFactory.addBootstrap(xmlstream.STREAM_END_EVENT, self._disconnected)

        # Map addBootstrap and removeBootstrap to the underlying factory -- is this
        # right? I have no clue...but it'll work for now, until i can think about it
        # more.
        self.addBootstrap = self._xsFactory.addBootstrap
        self.removeBootstrap = self._xsFactory.removeBootstrap

    def getFactory(self):
        return self._xsFactory

    def _connected(self, xs):
        self.xmlstream = xs
        for c in self:
            if IService.providedBy(c):
                c.transportConnected(xs)

    def _authd(self, xs):
        # Flush all pending packets
        for p in self._packetQueue:
            self.xmlstream.send(p)
        self._packetQueue = []
        # Notify all child services which implement
        # the IService interface
        for c in self:
            
            if IService.providedBy(c):
                c.componentConnected(xs)

    def _disconnected(self, _):
        self.xmlstream = None

        # Notify all child services which implement
        # the IService interface
        for c in self:
            if IService.providedBy(c):
                c.componentDisconnected()

    def send(self, obj):
        if self.xmlstream != None:
            self.xmlstream.send(obj)
        else:
            self._packetQueue.append(obj)


class LogService(Service):

    def transportConnected(self, xmlstream):
        xmlstream.rawDataInFn = self.rawDataIn
        xmlstream.rawDataOutFn = self.rawDataOut

    def rawDataIn(self, buf):
        print "RECV: %s" % unicode(buf, 'utf-8').encode('ascii', 'replace')

    def rawDataOut(self, buf):
        print "SEND: %s" % unicode(buf, 'utf-8').encode('ascii', 'replace')
        



def buildServiceManager(jid, password, strport, id = None):
    """ Constructs a pre-built L{ServiceManager}, using the specified strport
        string.    
    """
    svc = ServiceManager(jid, password, id)
    client_svc = jstrports.client(strport, svc.getFactory())
    client_svc.setServiceParent(svc)
    return svc
